
#include-once
#include '_A3D_Header.au3'
#include '_A3D_Camera.au3'
#include '_A3D_Transform.au3'
#include '_A3D_Shape.au3'

#include 'Include\Vec3f.au3'
#include 'Include\Matrix.au3'
#include 'Include\Sort.au3'
#include 'Include\NumericLimits.au3'

#include <Array.au3>
#include <GDIPlus.au3>

;~ Global $graphics

; Men of culture
Opt('MustDeclareVars', 1)

; Auslagern von Stuff der in jeden Frame gebraucht wird.
Global $A3D_PEN = 0
Global $A3D_BRU = 0
Global $A3D_MAT = 0
Global $A3D_POI = 0 ; Pointer of PolygonStruct
Global $A3D_PPT = 0

; Viewport settings inside Startup
Func _A3D_Startup($iWidth = 600, $iHeight = 450, $nFovDeg = 70, $nClipNear = 0.1, $nClipFar = 100)
	GUIRegisterMsg(522, __A3D_MouseWheel) ; $WM_MOUSEWHEEL = 522
	__A3D_CameraStartup()
	_GDIPlus_Startup()
	If @NumParams = 1 Then $iHeight = Int($iWidth / 1.3)
	__A3D_UpdateViewport($iWidth, $iHeight, $nFovDeg, $nClipNear, $nClipFar)
	If $A3D_PEN = 0 Then
		$A3D_PEN = _GDIPlus_PenCreate($A3G_DEFAULT_COLOR, 2)
		_GDIPlus_PenSetLineJoin($A3D_PEN, 2)
	EndIf
	If $A3D_BRU = 0 Then $A3D_BRU = _GDIPlus_BrushCreateSolid($A3G_DEFAULT_COLOR)
	If $A3D_MAT = 0 Then $A3D_MAT = DllCall($__g_hGDIPDll, 'int', 'GdipCreateMatrix2', 'float', $A3D_WIDTH / 2, 'float', 0, 'float', 0, 'float', $A3D_HEIGHT / 2, 'float', $A3D_WIDTH / 2, 'float', $A3D_HEIGHT / 2, 'handle*', 0)[7]
	If Not IsDllStruct($A3D_POI) Then
		$A3D_POI = DllStructCreate('float x1;float y1;float x2;float y2;float x3;float y3;float x4;float y4;float x5;float y5;')
		$A3D_PPT = DllStructGetPtr($A3D_POI)
	EndIf
EndFunc

; Aufräumdienst
Func _A3D_Shutdown()
	__A3D_CameraShutdown()
	If $A3D_PEN <> 0 Then _GDIPlus_PenDispose($A3D_PEN)
	If $A3D_BRU <> 0 Then _GDIPlus_BrushDispose($A3D_BRU)
	If $A3D_MAT <> 0 Then _GDIPlus_MatrixDispose($A3D_MAT)
	_GDIPlus_Shutdown()
EndFunc


Func _A3D_BeginScene()

	; Counter auf 0 setzen (keine Vertices und keine Indices vorhanden)
	$A3D_VC = 0
	$A3D_IC = 0

	; CameraMatrix und CameraPos berechnen
	__A3D_UpdateCamera()

	; FrameMatrix aus ProjectionMatrix & CameraMatrix berechnen
	$A3D_FM = M44MulM44($A3D_PM, $A3D_CM)

EndFunc

Func _A3D_GeometryDraw(ByRef $aGeo)
	__A3D_GeometryDraw($aGeo[0], $aGeo[1], $aGeo[2], $aGeo[4])
EndFunc

; Das Ende
Func _A3D_EndScene($hGFX)

	; z-Sorting
	Local $z = 0
	For $i = 0 To $A3D_IC - 1 Step 1
		$z = 0
		Local $aObj = $A3D_IA[$i][0]
		For $j = 1 To $aObj[0] Step 1
			$z += ($A3D_VA[$aObj[$j]])[2]
		Next
		$A3D_ZA[$i][0] = -$z / $aObj[0] ; center z (für punkte linien und alle flachen polygone ist das optimierbar!) lohnt sich aber aktuell nicht da andere Sachen langsamer sind.
		$A3D_ZA[$i][1] = $i
	Next
	If $A3D_IC > 1 Then __A3DQS($A3D_ZA, 0, $A3D_IC - 1) ; innerer Teil von Quicksort. Für "schnelle" ansichten __A3DQS2 nutzen (ist nicht "richtig" sortiert, dafür viel schneller)

	DllCall($__g_hGDIPDll, 'int', 'GdipSetWorldTransform', 'handle', $hGFX, 'handle', $A3D_MAT)

	Local $iObj, $aObj, $aNormal, $nAngle
	For $i = 0 To $A3D_IC - 1 Step 1
		$iObj = $A3D_ZA[$i][1]
		$aObj = $A3D_IA[$iObj][0]
		Switch $aObj[0]
			Case 1
				; Pointsize ?
				DllCall($__g_hGDIPDll, 'int', 'GdipSetSolidFillColor', 'handle', $A3D_BRU, 'dword', $A3D_IA[$iObj][1])
				DllCall($__g_hGDIPDll, 'int', 'GdipFillEllipse', 'handle', $hGFX, 'handle', $A3D_BRU, 'float', ($A3D_VA[$aObj[1]])[0] - 0.01, 'float', ($A3D_VA[$aObj[1]])[1] - 0.01, 'float', 0.02, 'float', 0.02)
			Case 2
				; Linewidth ?
				DllCall($__g_hGDIPDll, 'int', 'GdipSetPenColor', 'handle', $A3D_PEN, 'dword', $A3D_IA[$iObj][1])
				DllCall($__g_hGDIPDll, 'int', 'GdipDrawLine', 'handle', $hGFX, 'handle', $A3D_PEN, 'float', ($A3D_VA[$aObj[1]])[0], 'float', ($A3D_VA[$aObj[1]])[1], 'float', ($A3D_VA[$aObj[2]])[0], 'float', ($A3D_VA[$aObj[2]])[1])
			Case 3
				DllCall($__g_hGDIPDll, 'int', 'GdipSetSolidFillColor', 'handle', $A3D_BRU, 'dword', $A3D_IA[$iObj][1])
				$A3D_POI.x1 = ($A3D_VA[$aObj[1]])[0]
				$A3D_POI.y1 = ($A3D_VA[$aObj[1]])[1]
				$A3D_POI.x2 = ($A3D_VA[$aObj[2]])[0]
				$A3D_POI.y2 = ($A3D_VA[$aObj[2]])[1]
				$A3D_POI.x3 = ($A3D_VA[$aObj[3]])[0]
				$A3D_POI.y3 = ($A3D_VA[$aObj[3]])[1]
				DllCall($__g_hGDIPDll, 'int', 'GdipFillPolygon', 'handle', $hGFX, 'handle', $A3D_BRU, 'struct*', $A3D_PPT, 'int', 3, 'int', 'FillModeAlternate')
			Case 4
				DllCall($__g_hGDIPDll, 'int', 'GdipSetSolidFillColor', 'handle', $A3D_BRU, 'dword', $A3D_IA[$iObj][1])
				$A3D_POI.x1 = ($A3D_VA[$aObj[1]])[0]
				$A3D_POI.y1 = ($A3D_VA[$aObj[1]])[1]
				$A3D_POI.x2 = ($A3D_VA[$aObj[2]])[0]
				$A3D_POI.y2 = ($A3D_VA[$aObj[2]])[1]
				$A3D_POI.x3 = ($A3D_VA[$aObj[3]])[0]
				$A3D_POI.y3 = ($A3D_VA[$aObj[3]])[1]
				$A3D_POI.x4 = ($A3D_VA[$aObj[4]])[0]
				$A3D_POI.y4 = ($A3D_VA[$aObj[4]])[1]
				DllCall($__g_hGDIPDll, 'int', 'GdipFillPolygon', 'handle', $hGFX, 'handle', $A3D_BRU, 'struct*', $A3D_PPT, 'int', 4, 'int', 'FillModeAlternate')
			Case 5
				DllCall($__g_hGDIPDll, 'int', 'GdipSetSolidFillColor', 'handle', $A3D_BRU, 'dword', $A3D_IA[$iObj][1])
				$A3D_POI.x1 = ($A3D_VA[$aObj[1]])[0]
				$A3D_POI.y1 = ($A3D_VA[$aObj[1]])[1]
				$A3D_POI.x2 = ($A3D_VA[$aObj[2]])[0]
				$A3D_POI.y2 = ($A3D_VA[$aObj[2]])[1]
				$A3D_POI.x3 = ($A3D_VA[$aObj[3]])[0]
				$A3D_POI.y3 = ($A3D_VA[$aObj[3]])[1]
				$A3D_POI.x4 = ($A3D_VA[$aObj[4]])[0]
				$A3D_POI.y4 = ($A3D_VA[$aObj[4]])[1]
				$A3D_POI.x5 = ($A3D_VA[$aObj[5]])[0]
				$A3D_POI.y5 = ($A3D_VA[$aObj[5]])[1]
				DllCall($__g_hGDIPDll, 'int', 'GdipFillPolygon', 'handle', $hGFX, 'handle', $A3D_BRU, 'struct*', $A3D_PPT, 'int', 5, 'int', 'FillModeAlternate')
		EndSwitch

	Next
	DllCall($__g_hGDIPDll, 'int', 'GdipResetWorldTransform', 'handle', $hGFX)
EndFunc

; Culling1a:
; Überprüft ob sich das gesamte Objekt geometrisch hinter der Kamera befindet.
; FOV und Aspect unabhängig, daher lieber einen großen Winkel wählen.
Func __A3D_Cull1a(ByRef $aMinMax)
	Local $worldVertexCoord, $vtxvektor
	Local $normalizedCam = Vec3fNorm($A3D_CP)
	Local $aCenter = [$aMinMax[0] + $aMinMax[3]/2, $aMinMax[1] + $aMinMax[4]/2, $aMinMax[2] + $aMinMax[5]/2]
	$worldVertexCoord = M44MulXYZV($A3D_DM, $aCenter)
	$vtxvektor = V3Sub($A3D_CP, $worldVertexCoord)
	Vec3fRefNorm($vtxvektor)
	Local $scalar = $vtxvektor[0] * $normalizedCam[0] + $vtxvektor[1] * $normalizedCam[1] + $vtxvektor[2] * $normalizedCam[2]
	Return $scalar < 0.4 ; Bis zu einem Aspect von 2:1 ist 0.4 eine gute Wahl. Object based culling ist extrem agressiv und unterscheidet nicht zwischen x und y. Daher: Schauen dass der aspect passt.
EndFunc

; Culling1b:
; Überprüft ob sich das gesamte Objekt außerhalb des Viewports befindet.
; Deutlich genauer als Culling1, aber nicht sensitiv auf Spiegelungen (also Objekte hinter der Kamera), daher NACH Cull1a aufrufen.
Func __A3D_Cull1b(ByRef $aMinMax, ByRef $aTrafo)

	; TODO -> Mein erster Versuch war eine Kugel zu transformieren (daher Center + Radius, was vorher als BoundingSphere gespeichert war)
	; Es ist aber wohl besser einen Würfel mit 8 Punkten zu transformieren und darüber zu bestimmen ob das Objekt sichtbar ist.
	; ggf. auch ein Tetraeder dessen Inkugel die BoundingSphere ist (dann sinds nur 4 Punkte pro Objekt), solange man die Kugel
	; nach der Abbildung in den Screenspace noch herausfinden kann... bin nicht sicher.

	Local $aCenter = [$aMinMax[0] + $aMinMax[3]/2, $aMinMax[1] + $aMinMax[4]/2, $aMinMax[2] + $aMinMax[5]/2]
	Local $nRadius = Sqrt(($aMinMax[0] - $aMinMax[3])^2 + ($aMinMax[1] - $aMinMax[4])^2 + ($aMinMax[2] - $aMinMax[4])^2)


	Local $v = M44MulXYZV2($aTrafo, $aCenter)

	Local $dx = [$aCenter[0] + 0.0001, $aCenter[1], $aCenter[2]]
	Local $vdx = M44MulXYZV2($aTrafo, $dx)
	Local $sx = Sqrt(($v[0] - $vdx[0])^2 + ($v[1] - $vdx[1])^2) / 0.0001

	Local $dy = [$aCenter[0], $aCenter[1] + 0.0001, $aCenter[2]]
	Local $vdy = M44MulXYZV2($aTrafo, $dy)
	Local $sy = Sqrt(($v[0] - $vdy[0])^2 + ($v[1] - $vdy[1])^2) / 0.0001

	Local $dz = [$aCenter[0], $aCenter[1], $aCenter[2] + 0.0001]
	Local $vdz = M44MulXYZV2($aTrafo, $dz)
	Local $sz = Sqrt(($v[0] - $vdz[0])^2 + ($v[1] - $vdz[1])^2) / 0.0001

;~ 	ConsoleWrite($sx & ' ' & $sy & ' ' & $sz & @CRLF)

	Local $s = $sx
	If $sy > $s Then $s = $sy
	If $sz > $s Then $s = $sz
	; $s = Skalenfaktor in xyz richtung. Den muss man doch schlauer aus der Matrix holen können als auszuprobieren.
	; Eigenwerte wären eine Idee, das dauert aber zu lange.
	; Ein Problem hier ist, dass der Skalenfaktor auf diese Art berechnet quasi Eigenwerte entlang xyz sind, da
	; xyz aber keine Eigenvektoren der Matrix sind ist der Faktor hier ein Stück weit daneben... Ansich würde
	; uns der größte Eigenwert interessieren.
	; Das Ding hier ist ein TODO für später.

	Local $size = $nRadius * $s

	Local $clip = 1.0 + $size / 2
	Local $ret = $v[0] < -$clip Or $v[0] > $clip Or $v[1] < -$clip Or $v[1] > $clip

	; Debug Kreise um die Objekte um zu schauen wie die Boundingsphere transformiert aussieht (ist nur eine Approximation, siehe Text oben).
;~ 	If $ret Then
;~ 	DllCall($__g_hGDIPDll, 'int', 'GdipSetWorldTransform', 'handle', $graphics, 'handle', $A3D_MAT)

;~ 	DllCall($__g_hGDIPDll, 'int', 'GdipSetPenColor', 'handle', $A3D_PEN, 'dword', 0xFF8080FF)
;~ 	_GDIPlus_GraphicsDrawEllipse($graphics, $v[0] - $size/2, $v[1] - $size/2, $size, $size, $A3D_PEN)

;~ 	DllCall($__g_hGDIPDll, 'int', 'GdipResetWorldTransform', 'handle', $graphics)
;~ 	EndIf

	Return $ret


EndFunc

; Culling2:
; Überprüft per Objektelement, ob es sich außerhalb des Viewports befindet.
; Genauer als Culling2 aber deutlich rechenintensiver da deutlich häufiger aufgerufen (pro Objektelement)
Func __A3D_Cull2(ByRef $aElement)
	; Gab es mal, aber das sah wirklich nicht gut aus.
	; ggf. ist es auch schneller einfach "trotzdem zu zeichnen" selbst wenn
	; das Element teilweise/ganz außerhalb ist. Komplette Objekte sind ja eh weggecullt, wenn sie nicht sicherbar sind.
	; Wenn mir hier eine "schnelle" Methode einfällt (z.B. Skalenfaktor + Mittelpunkt etc) dann lohnt sich das vllt, aktuell ist es zu lahm und ich hab keine Ideen.
EndFunc

; Culling2:
; Backfaceculling - Für alles was nach Culling 1a, 1b, 2 übrig bleibt.
Func __A3D_Cull3(ByRef $a, ByRef $b, ByRef $c) ; im Prinzip -> trianglenormal dot cameranormal -> wenn <= 0 -> raus hier.
	Return (($a[1] - $b[1]) * ($a[2] - $c[2]) - ($a[2] - $b[2]) * ($a[1] - $c[1])) * $A3D_CN[0] + (($a[2] - $b[2]) * ($a[0] - $c[0]) - ($a[0] - $b[0]) * ($a[2] - $c[2])) * $A3D_CN[1] + (($a[0] - $b[0]) * ($a[1] - $c[1]) - ($a[1] - $b[1]) * ($a[0] - $c[0])) * $A3D_CN[2] <= 0
EndFunc

Func __A3D_GeometryDraw(ByRef $aVertices, ByRef $aIndices, ByRef $aColors, ByRef $aMinMax)
	Local $_ = $A3D_VC, $a, $m = M44MulM44($A3D_FM, $A3D_DM), $v

	If __A3D_Cull1a($aMinMax) Then Return
	If __A3D_Cull1b($aMinMax, $m) Then Return ; Skalenfaktor vorher berechnen?

	For $i = 0 To UBound($aVertices) - 1 Step 1
		$v = M44MulXYZV($m, $aVertices[$i])
		$A3D_VA[$_ + $i] = $v[2] > $A3D_CLIPNEAR And $v[2] < $A3D_CLIPFAR ? $v : 1 ; Das macht nicht wirklich was es soll... Hilft aber gegen murks beim rendern.
	Next

	For $i = 0 To UBound($aIndices) - 1 Step 1
		$a = $aIndices[$i]
		Switch $a[0]
			Case 1
				$a[1] += $_
				If Not IsArray($A3D_VA[$a[1]]) Then ContinueLoop ; klappt weil -> Arrays zählen als false, die 1 zählt als true. EDIT: Klappt wohl nicht immer. TODO
			Case 2
				$a[1] += $_
				$a[2] += $_
				If Not IsArray($A3D_VA[$a[1]]) Or Not IsArray($A3D_VA[$a[2]]) Then ContinueLoop
			Case 3
				$a[1] += $_
				$a[2] += $_
				$a[3] += $_
				If Not IsArray($A3D_VA[$a[1]]) Or Not IsArray($A3D_VA[$a[2]]) Or Not IsArray($A3D_VA[$a[3]]) Then ContinueLoop
				If __A3D_Cull3($A3D_VA[$a[1]], $A3D_VA[$a[2]], $A3D_VA[$a[3]]) Then ContinueLoop
			Case 4
				$a[1] += $_
				$a[2] += $_
				$a[3] += $_
				$a[4] += $_
				If Not IsArray($A3D_VA[$a[1]]) Or Not IsArray($A3D_VA[$a[2]]) Or Not IsArray($A3D_VA[$a[3]]) Or Not IsArray($A3D_VA[$a[4]]) Then ContinueLoop
				If __A3D_Cull3($A3D_VA[$a[1]], $A3D_VA[$a[2]], $A3D_VA[$a[3]]) Then ContinueLoop
			Case 5
				$a[1] += $_
				$a[2] += $_
				$a[3] += $_
				$a[4] += $_
				$a[5] += $_
				If Not IsArray($A3D_VA[$a[1]]) Or Not IsArray($A3D_VA[$a[2]]) Or Not IsArray($A3D_VA[$a[3]]) Or Not IsArray($A3D_VA[$a[4]]) Or Not IsArray($A3D_VA[$a[5]]) Then ContinueLoop
				If __A3D_Cull3($A3D_VA[$a[1]], $A3D_VA[$a[2]], $A3D_VA[$a[3]]) Then ContinueLoop
		EndSwitch

		$A3D_IA[$A3D_IC][0] = $a
		$A3D_IA[$A3D_IC][1] = $aColors[$i]
		$A3D_IC += 1

	Next

	$A3D_VC += UBound($aVertices)
EndFunc

Func __A3D_MouseWheel($hWnd, $iMsg, $wParam, $lParam)
	Local $p = BitShift($wParam, 16) > 0 ? -0.091 : 0.1
	_A3D_CameraZoom(1 + $p)
EndFunc
