#include-once
#include 'Include\Vec3f.au3'

; Author: M

; Wenn man mit Vertices arbeitet ist eine gridbasierte Beschleunigungsstruktur hilfreich.
; Damit sind Abfragen wie "Kann ich hier einen Vertex hinzufgen schneller als O(n) machbar
; eigentlich msste man den hinzuzufgenden Vertex gegen "alle" bisher vorhandenen gegenchecken.

; Methode:
; Skaliere alle Vertices in einen bereich von [0, iSize].
; 3D Array mit [iSize][iSize][iSize] enthlt jeweils eine Map
; Fge Vertices in die jeweils passende Map hinzu.

Global Const $LUT_EPS = 1e-9 ; Alles was nher beisammen ist als dieser Wert wird als "ein" Ort gewertet.

Func _VLut_Create(ByRef $aVertices, ByRef $aMinMax, ByRef $aScaleBox, ByRef $iVertexCount)
	Local $iSize = Round(2^(Log(UBound($aVertices))/Log(10)-1) + 1) ; Jop... da msste halbwegs sinnvoll sein. Je grer die Anzahl Vertices, desto groer der LUT

	__VLut_GeometryRescale($aVertices, $aMinMax, $aScaleBox, $iSize)

	Local $m[], $aXYZ[$iSize][$iSize][$iSize]
	For $z = 0 To $iSize - 1 Step 1
		For $y = 0 To $iSize - 1 Step 1
			For $x = 0 To $iSize - 1 Step 1
				$aXYZ[$x][$y][$z] = $m
			Next
		Next
	Next
	For $i = 0 To UBound($aVertices) - 1 Step 1 ; Alles hinzugfen (ohne Check obs schon drin ist. Wie nehmen an hier ist alles richtig)
		__VLut_Insert($aXYZ, $aVertices[$i], $iVertexCount)
	Next
	Return $aXYZ
EndFunc

Func _VLut_GetAll(ByRef $aLut, ByRef $aScaleBox, ByRef $aMinMax, ByRef $iVertexCount)
	Local $aVertices[$iVertexCount], $iSize = UBound($aLut), $List
	For $x = 0 To $iSize - 1 Step 1
		For $y = 0 To $iSize - 1 Step 1
			For $z = 0 To $iSize - 1 Step 1
				$List = $aLut[$x][$y][$z]
				For $i = 0 To UBound($List) - 1 Step 1 ; Lang ists her, dass ein 4x Loop sinnvoll war, aber hier gehts nicht anders.
					Local $a = [$List[$i][0], $List[$i][1], $List[$i][2]]
					$aVertices[$List[$i][3]] = $a
				Next
			Next
		Next
	Next
	__VLut_GeometryRescale($aVertices, $aMinMax, $aScaleBox)
	Return $aVertices
EndFunc

Func _VLut_Find(ByRef $aLut, ByRef $v, ByRef $iVertexCount)
	Local $List = $aLut[$v[0]][$v[1]][$v[2]]
	For $i = 0 To UBound($List) - 1 Step 1
		If ($v[0] - $List[$i][0])^2 + ($v[1] - $List[$i][1])^2 + ($v[2] - $List[$i][2])^2 < $LUT_EPS Then Return ($List[$i])[3]
	Next
	Return __VLut_Insert($aLut, $v, $iVertexCount)
EndFunc



; direkter Insert
Func __VLut_Insert(ByRef $aLut, $v, ByRef $iVertexCount)
	Local $m[] ; implizite Mapdeklaration fehlt in AutoIt... eieiei
	$m[0] = $v[0]
	$m[1] = $v[1]
	$m[2] = $v[2]
	$m[3] = $iVertexCount
	MapAppend($aLut[$v[0]][$v[1]][$v[2]], $m)
	$iVertexCount += 1
	Return $iVertexCount - 1
EndFunc

; Skalierung auf [0, iSize]
Func __VLut_GeometryRescale(ByRef $aVertices, ByRef $aMinMax, ByRef $aScaleBox, $iSize = 0)
	If IsArray($aScaleBox) Then ; Skaliert das Objekt so dass die Werte bei 0 beginnen und bei $iSize - 1 enden
		Local $aPos = [$aScaleBox[0], $aScaleBox[1], $aScaleBox[2]]
		Local $aScale = [$aScaleBox[3], $aScaleBox[4], $aScaleBox[5]]
		For $i = 0 To UBound($aVertices) - 1 Step 1 ; Zurck abbilden
			Vec3fRefMul($aVertices[$i], $aScale)
			Vec3fRefAdd($aVertices[$i], $aPos)
		Next
	Else
		Local $aPos = [$aMinMax[0], $aMinMax[1], $aMinMax[2]]
		Local $aScale = [$aMinMax[3] - $aMinMax[0], $aMinMax[4] - $aMinMax[1], $aMinMax[5] - $aMinMax[2]]
		Vec3fRefScalarDiv($aScale, $iSize - 1)
		Local $aTmp = [$aPos[0], $aPos[1], $aPos[2], $aScale[0], $aScale[1], $aScale[2]]
		$aScaleBox = $aTmp
		For $i = 0 To UBound($aVertices) - 1 Step 1 ; Abbildung auf Einheitswrfel
			Vec3fRefSub($aVertices[$i], $aPos)
			Vec3fRefDiv($aVertices[$i], $aScale)
		Next
	EndIf
EndFunc
